
#define	ACE_COFF 1
#include <stdio.h>
#include "filehdr.h"
#include "aouthdr.h"
#include "scnhdr.h"
#include "reloc.h"

#include "syms.h"
#include "ldfcn.h"
#include "crt.h"


typedef enum { True = 1, False = 0 }Boolean;

#define ERROR( error_msg )	{				\
		printf error_msg ;				\
		exit(-1);					\
}


#define DEBUG
#ifdef DEBUG
#define DEBUG_INFO( msg )	printf msg
#else
#define DEBUG_INFO( msg )
#endif

#define TEXT_BUFFER_SIZE	1024
unsigned char TextBuffer[TEXT_BUFFER_SIZE];

void Relocate( FILE *fp, long ptr, int size )
{
	int i;							
        long cur;						
	struct reloc reloc;					
	cur = ftell(fp);					
	for ( i = 0; i < size; i++ ){				
		if ( fread( &reloc, RELSZ, 1, fp )!= 1)
			ERROR(( "Could not read relocation entry\n" ));
		/* Section index starts from one, decrement to index from zero */
		reloc.r_symndx = reloc.r_symndx -1;
		if ( reloc.r_type != RSEG_LONG )
			continue;
		DEBUG_INFO(( "Relocation required\n" ));
#if 0
		*(unsigned long *)(&LoadBuf[reloc.r_vaddr]) += SectionAddress[reloc.r_symndx];
#endif
	}							
	fseek( fp, cur, 0 );					
}


void CopyTextSection( char *filename, FILE *fp, int NumSections, FILE *bfp )
{
	int i;
	SCNHDR SectionHeader;
	unsigned char *addr;
	long cur_pos;
	long total_size = 0;
	long buf_size, text_size;

	for ( i = 0; i < NumSections; i++ ){
		if ( fread( &SectionHeader, SCNHSZ, 1, fp ) != 1 )
			ERROR(( "%s: Could not read section header number %d\n", filename, i ));
		cur_pos = ftell(fp);
		if ( strcmp( SectionHeader.s_name, ".text" ) != 0 )
			continue;
		fseek( fp, SectionHeader.s_scnptr, 0 );
		text_size = SectionHeader.s_size;
		while ( text_size > 0 ){
			buf_size = ( text_size > TEXT_BUFFER_SIZE )?TEXT_BUFFER_SIZE:text_size;
			text_size -= buf_size;
			if ( fread((char *)TextBuffer,  sizeof(char), buf_size, fp ) != buf_size )
				ERROR(( "%s: Error in reading section %d (%s)\n", filename, i, SectionHeader.s_name ));
			if ( fwrite((char *)TextBuffer,  sizeof(char), buf_size, bfp ) != buf_size )
				ERROR(( "%s: Error in writing section %d (%s)\n", filename, i, SectionHeader.s_name ));
		}
		Relocate( fp, SectionHeader.s_relptr, SectionHeader.s_nreloc );
		fseek( fp, cur_pos, 0 );
	}
}


void CopyText( char *filename, char *binfilename )
{
	FILE *fp;
	FILE *bfp;
	FILHDR FileHeader;
	AOUTHDR	AoutHeader;
	int res;

	fp = fopen( filename, "r" );
	if ( fp == NULL  )
		ERROR(( "Could not open file %s\n", filename ));
	if (( res = fread( &FileHeader,  FILHSZ, 1, fp )) != 1 )
		ERROR(( "%s: Could not read file header, res %d\n", filename, res ));
	if (( FileHeader.f_magic != ACET800RG )&&( FileHeader.f_magic != ACET9000RG ))
		ERROR(( "%s: Bad magic %x\n", filename, FileHeader.f_magic ));
	if ( fread( &AoutHeader, AOUTHSZ, 1, fp ) != 1 )
		ERROR(( "%s: Could not read optional header\n", filename ));
	bfp = fopen( binfilename, "w" );
	CopyTextSection( filename, fp, (int)FileHeader.f_nscns,bfp );
	fclose( fp );
	fclose( bfp );
}


main( int argc, char **argv )
{
	if ( argc != 3 ){
		ERROR(( "Usage: copytext <COFF_file> <bin_file>\n" ));
	}
	CopyText( argv[1], argv[2] );
	exit(0);
}


