#include <stdio.h>
#include <assert.h>
#include <conc.h>
#include <sys/crt.h>
#include <node.h>
#include "OS.h"

#define SECONDARY_WORKSPACE_SIZE	144
#pragma PACT stackcheck off


/*	Fill the Process Info structure
*/
#define FILL_PROCESS_INFO( ProcessInfo, DataLo, DataSize, TextLo, TextSize,\
StackLo, StackSize, HeapLo, HeapSize, Entry ) {				\
	ProcessInfo->id = 0;						\
	ProcessInfo->mode = 3000;					\
	ProcessInfo->codesize = TextSize;				\
	ProcessInfo->code = (void *)TextLo;				\
	ProcessInfo->entry = (void *)Entry;				\
	ProcessInfo->stacksize = StackSize;				\
	ProcessInfo->stack = (void *)StackLo;				\
	ProcessInfo->vectorsize = 0;					\
	ProcessInfo->vector = NULL;					\
	ProcessInfo->datasize = DataSize;				\
	ProcessInfo->data = (void *)DataLo;				\
	ProcessInfo->heapsize = HeapSize;				\
	ProcessInfo->heap = (void *)HeapLo;				\
	ProcessInfo->next = (struct _Process *)NULL;			\
	ProcessInfo->nofparms = 2;					\
	ProcessInfo->nofvalues = 0;					\
}

/*	Fill the Node Info structure
*/
#define FILL_NODE_INFO( NodeInfo, BootOutChan, BootInChan, MemStart ) {	\
	NodeInfo->id = 1;						\
	NodeInfo->toserver = BootOutChan;				\
	NodeInfo->fromserver = BootInChan;				\
	NodeInfo->serversem.value = 1;					\
	NodeInfo->serversem.sf = NodeInfo->serversem.sb = 0;		\
	NodeInfo->memstart = MemStart;					\
	NodeInfo->devtype = 805;					\
	NodeInfo->devid = 11;						\
	NodeInfo->cpus = 0x7ffb;					\
	NodeInfo->host = 6;						\
	NodeInfo->server = 15;						\
	NodeInfo->OS = 4;						\
	NodeInfo->board = 0;						\
}

/*	Fill the ProcessorInfo structure
*/
#define FILL_PROCESSOR_INFO( ProcessorInfo, ProcessInfo, MemStart ){	\
	ProcessorInfo->id = 1;						\
	ProcessorInfo->mode = 0;					\
	ProcessorInfo->memsize = 0x400000;				\
	ProcessorInfo->extch = 0x80000000;				\
	ProcessorInfo->code = ProcessorInfo->data = MemStart;		\
	ProcessorInfo->linkin = 4;					\
	ProcessorInfo->linkout = 0;					\
	ProcessorInfo->freemem = NULL;					\
	ProcessorInfo->uprocs = ProcessInfo;				\
	ProcessorInfo->iprocs = NULL;					\
	ProcessorInfo->virdata = NULL;					\
	ProcessorInfo->nofchans = 0;					\
}

extern void CallEntryPoint( int (*fn)(), unsigned long *mdl, unsigned long *ParameterBlock );
extern void Load( void * address, Channel * BootInChan, int size );

/*	COFF loader
*/
CoffLoader( unsigned long *Links, Channel * BootInChan, Channel * BootOutChan, unsigned long MemStart )
{
	void *stackFragmentInfo[3];
	int *overflow;
	void *MDL[16];
	struct coff_params {
		unsigned long TextStart;
		unsigned long TextSize;
		unsigned long DataStart;
		unsigned long DataSize;
		unsigned long BssStart;
		unsigned long BssSize;
		unsigned long NumSections;
		int (*entry)();
		int HeapSize;
		int StackSize;
		int DebugFlag;
	}CoffParameters;
	struct section_info {
		unsigned long SectionStart;
		unsigned long SectionSize;
	}SectionInfo;
	int i;
	unsigned long ParameterBlock[3];
	struct _Node *_NodePtr;
	struct _Node _NodeInformation;
	struct _Process ProcessInfo;
	struct _Processor ProcessorInfo;
	int *NewStack;
	int *ptr;

	Load( &CoffParameters, BootInChan, sizeof(CoffParameters) );
	for ( i = 0; i < ((CoffParameters.BssSize > 0)?CoffParameters.NumSections-1:CoffParameters.NumSections); i++ ){
		Load( &SectionInfo, BootInChan, sizeof(SectionInfo) );
		if ( SectionInfo.SectionSize > 0 )
			Load( SectionInfo.SectionStart, BootInChan, SectionInfo.SectionSize );
	}
	ptr = (int *)CoffParameters.BssStart;
	for ( i = 0; i < CoffParameters.BssSize/sizeof(int); i++ )
		*ptr++ = 0;
	MDL[SYS_mdl_Magic] = (void *)SYS_MAGIC_MDL;
	MDL[SYS_mdl_TextLo] = CoffParameters.TextStart;
	MDL[SYS_mdl_TextHi] = (void*)((char*) MDL[SYS_mdl_TextLo] + CoffParameters.TextSize );
	MDL[SYS_mdl_DataLo] = CoffParameters.DataStart;
	MDL[SYS_mdl_DataHi] = (void*)((char*) MDL[SYS_mdl_DataLo] + CoffParameters.DataSize + CoffParameters.BssSize );
	MDL[OS_mdl_HeapLo] = MDL[SYS_mdl_DataHi] > MDL[SYS_mdl_TextHi]?(void*)((char*) MDL[SYS_mdl_DataHi]):MDL[SYS_mdl_TextHi];
	MDL[OS_mdl_HeapHi] = (void*)((char*) MDL[OS_mdl_HeapLo] + CoffParameters.HeapSize );

	NewStack = (( int *)(MDL[OS_mdl_HeapHi]))+(CoffParameters.StackSize/4 - SECONDARY_WORKSPACE_SIZE - 10);
	overflow =  ((void * (*) (const char *)) __asm)( "ldlp 0" );
	for ( i = -5; i < SECONDARY_WORKSPACE_SIZE; i++ )
		NewStack[i] = overflow[i];
	__asm( "ldl NewStack$16" ); __asm( "gajw" );

	/* init the stack fragment info block: */
	stackFragmentInfo[1] = SYS_NULL_ptr;	/* Prev */
	stackFragmentInfo[2] = SYS_NULL_ptr;	/* Next */
	overflow =  ((void * (*) (const char *)) __asm)( "ldlp 0" );

	stackFragmentInfo[0] = MDL[SYS_mdl_StackLo] = (void*) (MDL[OS_mdl_HeapHi]);
	MDL[SYS_mdl_StackOvfl] = (void*) (((int *)MDL[OS_mdl_HeapHi])+80);

	MDL[SYS_mdl_StackHi] = (void*)	( &stackFragmentInfo[3] );
	MDL[SYS_mdl_StackClass] = 0;
	MDL[OS_link_to_server] = BootOutChan;

	_NodePtr = &_NodeInformation;
	FILL_NODE_INFO( _NodePtr, BootOutChan, BootInChan, MemStart );
	_NodePtr->process = &ProcessInfo;
	FILL_PROCESS_INFO( _NodePtr->process, 
		CoffParameters.DataStart, CoffParameters.DataSize+CoffParameters.BssSize, 
		CoffParameters.TextStart, CoffParameters.TextSize, 
		MDL[SYS_mdl_StackHi], CoffParameters.StackSize, 
		MDL[OS_mdl_HeapLo], CoffParameters.HeapSize, 
		CoffParameters.entry );
	_NodePtr->process->processor = &ProcessorInfo;
	FILL_PROCESSOR_INFO( _NodePtr->process->processor, _NodePtr->process, MemStart );
	ParameterBlock[0] = 2 * sizeof(long);
	ParameterBlock[1] = (unsigned long)_NodePtr;
	ParameterBlock[2] = CoffParameters.DebugFlag;


	ResetTimer();
	CallEntryPoint( CoffParameters.entry, MDL, &(ParameterBlock[1]) );
}

