
/*
 *
 *	Stripped (XPDS, T9, COFF only) crt0.c.
 *
 *	BASED ON: @(#)startup.c	1.35	94/05/10	(c) PACT
 *
 *	Copyright (c) 1991-1994 PACT
 *	All Rights Reserved
 *
 *	This copy is for private circulation only and may not be used
 *	in any other manner.
 *
 *	C runtime entry (startup) code.  Run as user process after network has
 *	been loaded, sets up C runtime environment before calling main().
 *
 */

#include	<stdio.h>
#include	<asm.h>
#include	<par.h>
#include	<conc.h>
#include	<node.h>
#include	<stdlib.h>
#include	<string.h>
#include	<setjmp.h>
#include	<sys/times.h>
#include	"trap.h"


/* No stack checking, of course.										*/
#pragma	PACT stackcheck off

#define MAX_COMMAND_LINE_LENGTH	512
#define MAX_DEBUG_COMMAND_LEN	100
#define MAX_STRING_LENGTH	256
#define MAXARG			50

/* A bool is an int.													*/
typedef	int				bool_t;
#define	FALSE			0
#define	TRUE			1

#ifndef	_C_Entry
#	define	_C_Entry	_CRT_enter
#endif


static char *_ParameterAddr;

#define iswhitespace(ch) ((ch == ' ') || (ch == '\t') || (ch == '\0') || (ch == '\n'))

/*	Fill the Node Info structure
*/
#define FILL_NODE_INFO( NodeInfo, MemStart, FromServer, NodeId ) {	\
	NodeInfo.id = NodeId;						\
	NodeInfo.toserver = (unsigned long)FromServer + 4;		\
	NodeInfo.fromserver = FromServer;				\
	NodeInfo.serversem.value = 1;					\
	NodeInfo.serversem.sf = NodeInfo.serversem.sb = 0;		\
	NodeInfo.memstart = MemStart;					\
	NodeInfo.devtype = 805;						\
	NodeInfo.devid = 11;						\
	NodeInfo.cpus = 0x7ffb;						\
	NodeInfo.host = 6;						\
	NodeInfo.server = 115;						\
	NodeInfo.OS = 4;						\
	NodeInfo.board = 0;						\
}

/*	Fill the ProcessorInfo structure
*/
#define FILL_PROCESSOR_INFO( ProcessorInfo, ProcessInfo, MemStart, MemSize ){	\
	ProcessorInfo.id = 1;						\
	ProcessorInfo.mode = 0;						\
	ProcessorInfo.memsize = MemSize;				\
	ProcessorInfo.extch = 0x80000000;				\
	ProcessorInfo.code = ProcessorInfo.data = MemStart;		\
	ProcessorInfo.linkin = 4;					\
	ProcessorInfo.linkout = 0;					\
	ProcessorInfo.freemem = NULL;					\
	ProcessorInfo.uprocs = ProcessInfo;				\
	ProcessorInfo.iprocs = NULL;					\
	ProcessorInfo.virdata = NULL;					\
	ProcessorInfo.nofchans = 0;					\
}
/*
 *
 *		void _C_Entry2 (struct _Process *proc)
 *
 *	This function is called by _C_Entry with a proper GP and PD.  It sets
 *	up the runtime environment and calls main().
 *
 */

int	_CRT_enter ( struct _Process *proc, int DebugFlag, char *DebugCommand, unsigned long FromServer, unsigned long NodeId, int MemSize, char *Params )
{
	/* Declare some functions and vars used: memory manager, file		*/
	/* system, debugger etc.											*/
	extern	void	*_mdesc, *__minit (void *heap, int heapsize);
	extern	void	_finit (void);
	extern	int		*_dbginit (void *code, void *data);
	extern	int		main (int argc, char *argv[], char *envp[]);

	/* Some functions use different linkage.							*/
	int ret, i;
	long length;
	int argc; char **argv; char **envp;
	void *MemStart;
	struct _Processor ProcessorInfo;
	struct _Parm *parms, *param;

#	pragma	PACT	linkage tcc	__minit

	/* Init memory manager.  Gets heap.									*/

	_mdesc = __minit (proc->heap, proc->heapsize);

	MemStart      = ((void * (*) (const char *)) __asm) ("ldmemstartval");
	_ParameterAddr = Params;
	FILL_NODE_INFO( _Node, MemStart, FromServer, NodeId );
	_Node.process = proc;
	_Node.process->processor = &ProcessorInfo;
	FILL_PROCESSOR_INFO( ProcessorInfo, _Node.process, MemStart, MemSize );


	/* Store pointer to start of Processor structure at start of 		*/
	/* memory, for debugging purposes.									*/
	((struct _Processor **) _Node.memstart) [0] = proc->processor;

	/* If not reduced, get host info, init filesystem and get cl args.	*/
/*	_serverid ();		940803,REHK: causes hangs later on! */
	_finit ();

	/*      Convert process to low priority */
	_ProcToLow();

	/* Setup longjmp env for tasks that want to return to the system	*/
	/* process that started them instead of stopping.					*/
/*	if (setjmp (_taskreturn_env) == 0)		*/

#ifdef GDBINIT
	SetArgcArgv( &argc, &argv );
	if ( DebugFlag )
		debug_init( DebugCommand, argv[0] );
	else
#endif
	SetupTrapHandler();
	envp = NULL;
	ret = main (argc, argv, envp);
#ifdef GDBINIT
	if ( DebugFlag )
		debug_exit();
#endif
	Exit(ret);
}

Exit( int retval )
{
	_LIBC_exit();
#ifdef GDBINIT
	__Exit(retval);
	__asm( "stopp" );
#else
	/* _exit causes the processor to hang: FIXME 
	   Other tasks are unable to run after _exit is called */
	_exit( retval );
#endif
}

void exit( int retval )
{
#ifdef GDBINIT
	debug_exit();
#endif
	Exit( retval );
}

static char *copystr(str)
char *str;
{
   char  *new;
   
   new = malloc(strlen(str)+1);
   if (new == NULL)
      return NULL;

   strcpy(new, str);
   
   return new;
}

#ifdef GDBINIT
static char argstr[MAX_COMMAND_LINE_LENGTH+2];
static char tempStr[MAX_STRING_LENGTH];
int SetArgcArgv(int *ArgcPtr, char **ArgvPtr[])
{
   static char *argv[MAXARG];
   int argc = 0;
   int i, j;
   int removed = 0;
   char *lp;
   char red,*rfname;
   char err;

   GetArgs( argstr );
   if (strlen(argstr) > MAX_STRING_LENGTH)
      return -1;
   argstr[strlen(argstr)+1] = '\0';
      
   lp = argstr;
   
   while (*lp)
   {
      /* find first character of this argument */
      while (iswhitespace(*lp))
         lp++;
         
      err = ' ';

      /* Check for I/O redirectors */
      if ((*lp == '<') || (*lp == '>')) {
         red = *lp++;
         if ((red == '>') && (*lp == '&'))
            err = *lp++;
            
         while (iswhitespace(*lp))
            lp++;
         
         rfname = lp;
         /* find the end of this argument */
         while (!iswhitespace(*lp))
            lp++;
         *lp++ = '\0';
         
      }
      else {
         /* Is it quoted ? */
         if (*lp == '"') {
            argv[argc++] = ++lp;
            
            /* Find the rest of the argument */
            while ((*lp != '"') && (*lp != '\0'))
               lp++;
         }
         else {
            argv[argc++] = lp;

            /* find the end of this argument */
            while (!iswhitespace(*lp))
               lp++;
         }

         *lp++ = '\0';
      }
   }
   
   for (i=0; i<argc; i++){
      strcpy( tempStr, argv[i] );
      for ( j = 0; j < strlen(tempStr); j++ )
         tempStr[j] = tolower(tempStr[j]);
      if (( strcmp( tempStr, "-d" ) == 0 )||
          ( strcmp( tempStr, "-heap" ) == 0 )||
          ( strcmp( tempStr, "-stack" ) == 0 )){
             for ( j = i+2; j <argc; j++ )
                argv[j-2] = argv[j];
             argc -= 2; 
             if ( i == argc) break;
             i -= 1;
      }
            
   }

   *ArgcPtr =  argc;
   *ArgvPtr = (char **)malloc( argc * sizeof(char *));
   for (i=0; i<argc; i++)
      (*ArgvPtr)[i] = copystr(argv[i]);
}
#endif


char *_get_param(int index)
{
	char *param;
	int i;
	if (( index < 1 )||( index > *_ParameterAddr ))
		return(0);
	param = _ParameterAddr+1;
	for ( i = 1; i < index; i++ )
		param += *param + 1;
	param++;
	return(param);
}

#if 1
void __semp( sem_t *sem )
{
	if ( sem->sf == 0 )sem->sf = 0x80000000;
	if ( sem->sb == 0 )sem->sb = 0x80000000;
	__mysemp( sem );
}
void __semv( sem_t *sem )
{
	if ( sem->sf == 0 )sem->sf = 0x80000000;
	if ( sem->sb == 0 )sem->sb = 0x80000000;
	__mysemv( sem );
}

void __seminit( sem_t *sem, int value )
{
	sem->value = value;
	sem->sf = sem->sb = 0x80000000;
}
#endif

static unsigned char TrapHandlerStack[TRAP_HANDLER_STACKSIZE];
extern void DefaultTrapHandler();

void SetupTrapHandler()
{
	THDS *th;
	int Wptr;

	Wptr = (int)(&TrapHandlerStack[TRAP_HANDLER_STACKSIZE-64-sizeof(THDS)]);
	GetMDL(Wptr+4);
	th = (THDS *)(&TrapHandlerStack[TRAP_HANDLER_STACKSIZE-sizeof(THDS)]);
	th->Fptr = th->Bptr = NOT_PROCESS_P;
	th->Iptr = (unsigned long)DefaultTrapHandler;
	th->eWu = th->eWl = 0;
	th->Cntl = IGNORE_ALL_TRAPS;
	SetTrapHandler( th );
}


