#include <stdio.h>
#include <conc.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <stdio.h>
#include <errno.h>

#define QLEN 10
#define TMPDIR	"."
#define DEBUG_TEMP_FILE	".gdbtmp"

static int DebuggerSocket = -1;
static int mySocket = -1;
static char *tempFileName = NULL;
static char DebugBuffer[BUFSIZ];
static int DebugBufCnt = 0;
static char *DebugBufPtr = DebugBuffer;

#undef ntohs
int debug_init( char *DebuggerCommand, char *DebugFileName )
{
	struct sockaddr_in name;
	struct protoent *protoent;
	int tmp;
	int sock, msock;
	int length;
	char Buf[100];
	FILE *fp;
	char *Debugger;

	sock = socket( AF_INET, SOCK_STREAM, 0 );
	if ( sock < 0 ){
		fprintf( stderr, "Cannot create socket for gdb, err %d\n", errno );
		return( -1 );
	}
	mySocket = sock;
	/* Allow rapid reuse of this port. */
	tmp = 1;
	setsockopt (sock, SOL_SOCKET, SO_REUSEADDR, (char *)&tmp, sizeof(tmp));

	/* Enable TCP keep alive process. */
	tmp = 1;
	setsockopt (sock, SOL_SOCKET, SO_KEEPALIVE, (char *)&tmp, sizeof(tmp));

	tmp = 1;
	if (setsockopt (sock, IPPROTO_TCP, TCP_NODELAY,
		  (char *)&tmp, sizeof(tmp)))
		return -1;


	bzero((char *)&name, sizeof(name));
	name.sin_family = AF_INET;
	name.sin_addr.s_addr = INADDR_ANY;
	name.sin_port = 0;
	if ( bind( sock, (struct sockaddr *)&name, sizeof(name)) < 0 ){
		fprintf( stderr, "Cannot bind socket to port for gdb, err %d\n", errno );
		return( -1 );
	}
        length = sizeof(name);
        if ( getsockname( sock, (struct sockaddr *)&name, &length) < 0 ){
                fprintf( stderr, "error getting socket name\n" );
                return(-1);
        }

	if ( listen(sock , QLEN ) < 0 ){
		fprintf( stderr, "Cannot listen to gdb port, err %d\n", errno );
		return( -1 );
	}
	gethostname( &(Buf[0]), 100 );
	tempFileName = DEBUG_TEMP_FILE;
	fp = fopen( tempFileName, "w" );
	if ( fp == NULL ){
		fprintf( stderr, "Cannot open temporary file\n" );
		return( -1 );
	}
	fprintf( fp, "file %s\n", DebugFileName );
	fprintf( fp, "target remote %s:%d\n", Buf,  ntohs(name.sin_port));
	fprintf( fp, "break main\n" );
	fprintf( fp, "continue\n" );
	fclose( fp );
	Debugger = (char *)malloc( strlen(DebuggerCommand)+strlen(tempFileName)+ 10 );
	sprintf( Debugger, "%s -command=%s &", DebuggerCommand, tempFileName );
	system( Debugger );
        length = sizeof(name);
        msock = accept( sock, (struct sockaddr *)&name, &length );
        if ( msock < 0 )
                fprintf( stderr, "Could not connect to gdb\n" );
	DebuggerSocket = msock;
	set_debug_traps();
	breakpoint();
        return ( msock );
}

debug_exit()
{
	if ( DebuggerSocket > 0 ){
		putpacket( "W00" );
		close( mySocket );
		close( DebuggerSocket );
		DebuggerSocket = -1;
		unlink( tempFileName );
		free( tempFileName );
	}
}



int putDebugChar(unsigned char c)
{
	return(write( DebuggerSocket, &c, 1 ));
}

int putDebugBuf(unsigned char *buf, int len)
{
	return(write( DebuggerSocket, buf, len ));
}

unsigned char getDebugChar()
{
	unsigned char c;
	getDebugBuf( &c, 1 );
	return(c);
}

int getDebugBuf( char *buf, int len )
{
	int count, i;
#if 0
	count = read( DebuggerSocket, buf, len );
	for ( i = 0; i < count; i++ )
		if ( buf[i] == 3 ) exit(-1);
	return( count );
#else
	if ( DebugBufCnt >= len ){
		DebugBufCnt -= len;
		memcpy( buf, DebugBufPtr, len );
		return(len);
	}
	DebugBufCnt = read( DebuggerSocket, DebugBuffer, BUFSIZ );
	if ( DebugBufCnt <= 0 )
		return(0);
	count = ( DebugBufCnt >= len )?len:DebugBufCnt;
	DebugBufCnt -= len;  DebugBufPtr = DebugBuffer;
	memcpy( buf, DebugBufPtr, count );
	DebugBufPtr += count;
	for ( i = 0; i < count; i++ )
		if ( buf[i] == 3 ) exit(-1);
	return(count);
#endif
}


flush_i_cache()
{}

