#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <netdb.h>
#include <errno.h>
#include <conc.h>
#include "tcl.h"
#include "tclInt.h"

#define HOST_NAME	"tcl_hostname"
#define PORT_NUMBER	"tcl_portnumber"
#define QLEN		10

/*
 * The following structure defines all of the commands supported by
 * VCP, and the C procedures that execute them.
 */

typedef struct {
    char *name;			/* Name of command. */
    int (*cmdProc) _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
	    int argc, char **argv));
				/* Command procedure. */
} VcpCmd;


VcpCmd VcpCommands[] = {
    /*
     * Commands that are part of the intrinsics:
     */

    {"run",		Tcl_RunCmd},
    {"vcpsend",		Tcl_SendCmd},
    {"vcprecv",		Tcl_RecvCmd},
    {"inspect",		Tcl_InspectCmd},
    {(char *) NULL,	(int (*)()) NULL}

};
#ifdef PVM
extern struct sockaddr_in TclSockAddr;
extern int TclSockNameLength;
extern int TclSocket;
extern void TclThread();

#undef ntohs
int TclOpenSocket( Tcl_Interp *interp )
{
	struct sockaddr_in name;
	int sock, msock;
	int length;
	char HostName[100], PortNumber[20];
	int i, j;

	sock = socket( AF_INET, SOCK_STREAM, 0 );
	if ( sock < 0 ){
		fprintf( stderr, "Cannot create socket for TCL, err %d\n", errno );
		return( -1 );
	}
	bzero((char *)&name, sizeof(name));
	name.sin_family = AF_INET;
	name.sin_addr.s_addr = INADDR_ANY;
	name.sin_port = 0;
	if ( bind( sock, (struct sockaddr *)&name, sizeof(name)) < 0 ){
		fprintf( stderr, "Cannot bind socket to port for TCL, err %d\n", errno );
		return( -1 );
	}
        length = sizeof(name);
        if ( getsockname( sock, (struct sockaddr *)&name, &length) < 0 ){
                fprintf( stderr, "error getting socket name\n" );
                return(-1);
        }

	if ( listen(sock , QLEN ) < 0 ){
		fprintf( stderr, "Cannot listen to TCL port, err %d\n", errno );
		return( -1 );
	}
	gethostname( HostName, 99 );
	sprintf( PortNumber, "%d", ntohs(name.sin_port));
	Tcl_SetVar(interp, HOST_NAME, HostName, TCL_GLOBAL_ONLY);
	Tcl_SetVar(interp, PORT_NUMBER, PortNumber, TCL_GLOBAL_ONLY);
	TclSockAddr = name;
        TclSockNameLength = sizeof(name);
	TclSocket = sock;

        return ( TclSocket );
}
#endif

Vcp_Init( Tcl_Interp *interp )
{
    /*
     * Bind in VCP's commands.
     */
    VcpCmd *cmdPtr;
    Process *TclProc;

    for (cmdPtr = VcpCommands; cmdPtr->name != NULL; cmdPtr++) {
	Tcl_CreateCommand(interp, cmdPtr->name, cmdPtr->cmdProc,
		(ClientData) 0, (void (*)()) NULL);
    }
    SetServerSemaphore(); /* Semaphore used by VCP and iserver */
#ifdef PVM
    TclSocket = TclOpenSocket(interp);
    TclProc = ProcAlloc( TclThread, 2048, 0 );
    ProcRunHigh( TclProc );
#endif
}
