/*{{{  includes*/


/* Copyright INMOS Limited 1988,1990 */

/* CMSIDENTIFIER */
static char *CMS_Id = "PRODUCT:ITEM.VARIANT-TYPE;0(DATE)";

#include <stdio.h>
#include <errno.h>
#include "inmos.h"
#include "isocket.h"
#include "vcpdefs.h"
#include "systype.h"
#include "queue.h"
#include "vcptype.h"
#include "iserver.h"
#define	ACE_COFF
#include <filehdr.h>
#include <aouthdr.h>
#include <scnhdr.h>
#include <reloc.h>
#include <sgs.h>
/*}}}*/

/*{{{  defines and globals*/

#define	BOOT_BUFFER_SIZE	(1024*10)
unsigned char BootBuffer[BOOT_BUFFER_SIZE];
unsigned char DebuggerCommand[MAX_COMMAND_LINE_LENGTH];
int AppHeapSize=100*1024, AppStackSize=100*1024;
int DebugApplicationSwitch=0;

/*  COFF parameters that are to be sent to the secondary bootstrap
*/
struct coff_params {
	unsigned long TextStart;
	unsigned long TextSize;
	unsigned long DataStart;
	unsigned long DataSize;
	unsigned long BssStart;
	unsigned long BssSize;
	unsigned long NumSections;
	unsigned long Entry;
	unsigned long HeapSize;
	unsigned long StackSize;
	unsigned long DebugFlag;
}CoffParams;

/*}}}*/

/*{{{  utilities*/

short _htons(short in)
{ short out;
  char *inptr=(char *)&in, *outptr=(char *)&out;
  outptr[0]=inptr[1];
  outptr[1]=inptr[0];
  return(out);
}

long _htonl(long in)
{ long out;
  char *inptr=(char *)&in, *outptr=(char *)&out;
  outptr[0]=inptr[3];
  outptr[1]=inptr[2];
  outptr[2]=inptr[1];
  outptr[3]=inptr[0];
  return(out);
}
/*}}}*/

/*{{{  Relocate*/
/*
 *------------------------------------------------------------------------
 *
 *	Relocate :
 *		Relocates a part of the section which has been read into
 *		BootBuffer.
 *	Parameters :
 *		fp : File pointer to open COFF file
 *		StartOffset : Offset from which to relocate
 *		EndOffset : Offset until which relocation is done
 *		RelPtr : Pointer to relocation info 
 *		nReloc : Pointer to number of relocation entries.
 *		LoadAddress : Address to which relocation is performed.
 *
 *	Return value :
 *		RelPtr is set to the file offset of the next relocation
 *		entry. nReloc is set to the number of relocation entries
 *		left. The function does not return anything.
 *
 *------------------------------------------------------------------------
 */
static void Relocate( fp, StartOffset, EndOffset, RelPtr, nReloc, LoadAddress, TheLink, ServerInfo )
FILE *fp;
unsigned long StartOffset;
unsigned long EndOffset;
long *RelPtr;
int *nReloc;
unsigned long LoadAddress;
int TheLink;
SERVER_INFO *ServerInfo;
{
	int i;
	long cur;
	struct reloc reloc;
	long CurRelPtr;
	int nRelocOver;
	unsigned long address;

	CurRelPtr = cur = ftell(fp);
	fseek( fp, *RelPtr, 0 );
	nRelocOver = 0;
	for ( i = 0; i < *nReloc; i++ ){
		CurRelPtr = ftell(fp);
		if ( fread( &reloc, RELSZ, 1, fp )!= 1)
			ABORT(MISC_EXIT, ( SE,"Could not read relocation entry" ));
		reloc.r_vaddr = _htonl( reloc.r_vaddr );		
		reloc.r_symndx = _htonl( reloc.r_symndx );	
		reloc.r_type = _htons( reloc.r_type );		
		/* Section index starts from one, decrement to index from zero */
		reloc.r_symndx--;
		if ( reloc.r_type != RSEG_LONG )
			continue;
		if ( reloc.r_vaddr < StartOffset ){
			ABORT(MISC_EXIT, ( SE,"Internal error: Unexpected reloc entry addr %x, off %x", reloc.r_vaddr,StartOffset ));
		}
		if ( reloc.r_vaddr >= EndOffset )
			break;
		address = *(unsigned long *)(&BootBuffer[reloc.r_vaddr-StartOffset]);
		address  += LoadAddress;
		*(unsigned long *)(&BootBuffer[reloc.r_vaddr-StartOffset]) = address;
		nRelocOver++;
	}
	*RelPtr = CurRelPtr;
	*nReloc -= nRelocOver;
	fseek( fp, cur, 0 );
}

/*}}}*/

/*{{{  LoadSections*/
/*
 *------------------------------------------------------------------------
 *
 *	LoadSections :
 *		Loads the sections of a COFF file at the specified address
 *		over a virtual or physical channel.
 *		CoffParameters are sent first. Each section is preceded by 
 *		its SectionInfo structure.
 *	Parameters :
 *		fp : File pointer to open COFF file
 *		NumSections : Number of sections in the COFF file
 *		LoadAddress : Address at which the file is loaded
 *		TheLink : Virtual/physical link number over which file is to
 *			be sent
 *		ServerInfo : structure containing information about server
 *
 *	Returns nothing.
 *
 *------------------------------------------------------------------------
 */
static void LoadSections( CoffFileName, fp, NumSections, TheLink, ServerInfo )
FILE *fp;
int NumSections;
int TheLink;
SERVER_INFO *ServerInfo;
{
	int i, j;
	unsigned long LoadAddress;
	SCNHDR SectionHeader;
	long cur_pos;
	struct section_info {
		unsigned long SectionStart;
		unsigned long SectionSize;
	}SectionInfo;
	int ops_res;
	int buf_size, section_size;
	int nreloc;
	unsigned long relptr, StartOffset, EndOffset;
	unsigned long MemRequired=0;

	DEBUG((  "Loading %d sections", NumSections ));
	cur_pos = ftell(fp);
	for ( i = 0; i < NumSections; i++ ){
		if ( fread( &SectionHeader, SCNHSZ, 1, fp ) != 1 )
			DEBUG(( "Could not read section header number %d", i ));
		SectionHeader.s_size = _htonl( SectionHeader.s_size );
		SectionHeader.s_paddr = _htonl( SectionHeader.s_paddr );
		SectionHeader.s_vaddr = _htonl( SectionHeader.s_vaddr );
		SectionHeader.s_scnptr = _htonl( SectionHeader.s_scnptr );
		SectionHeader.s_type = _htons( SectionHeader.s_type );
		SectionHeader.s_relptr = _htonl( SectionHeader.s_relptr );
		SectionHeader.s_nreloc = _htons( SectionHeader.s_nreloc );
		if ( strcmp( SectionHeader.s_name, ".text" ) == 0 ){
			CoffParams.TextStart = SectionHeader.s_vaddr;
			DEBUG(( "TextStart %x",  CoffParams.TextStart ));
			CoffParams.TextSize = SectionHeader.s_size;
		}
		if ( strcmp( SectionHeader.s_name, ".data" ) == 0 ){
			CoffParams.DataStart = SectionHeader.s_vaddr;
			CoffParams.DataSize = SectionHeader.s_size;
		}
		if ( strcmp( SectionHeader.s_name, ".bss" ) == 0 ){
			CoffParams.BssStart = SectionHeader.s_vaddr;
			CoffParams.BssSize = SectionHeader.s_size;
		}
	}
	MemRequired = CoffParams.TextSize+CoffParams.DataSize+CoffParams.BssSize+CoffParams.HeapSize+CoffParams.StackSize;
	if ((ops_res = WriteLink( TheLink, &MemRequired, sizeof(MemRequired), BOOT_TIMEOUT)) <= 0 )
		ABORT(MISC_EXIT, (SE,"Boot: Could not write COFF params, res %d\n", ops_res ));
	ReadLink( TheLink, &LoadAddress, sizeof(unsigned long), -1);
	DEBUG(( "MemStart = %x\n", LoadAddress ));
	CoffParams.TextStart += LoadAddress;
	CoffParams.DataStart += LoadAddress;
	CoffParams.BssStart += LoadAddress;
	CoffParams.Entry += LoadAddress;
	DEBUG(( "Writing Coff params Link %d size %d", TheLink, sizeof(CoffParams) ));
	if ((ops_res = WriteLink( TheLink, &CoffParams, sizeof(CoffParams), BOOT_TIMEOUT)) <= 0 )
		ABORT(MISC_EXIT, (SE,"Boot: Could not write COFF params, res %d\n", ops_res ));
	DEBUG(( "Wrote Coff params entry %x", CoffParams.Entry ));
	fseek( fp, cur_pos, 0 );
	for ( i = 0; i < NumSections; i++ ){
		if ( fread( &SectionHeader, SCNHSZ, 1, fp ) != 1 ){
			ABORT(MISC_EXIT, (SE, "Could not read section header number %d\n", i ));
		}
		cur_pos = ftell(fp);
		SectionHeader.s_size = _htonl( SectionHeader.s_size );
		SectionHeader.s_paddr = _htonl( SectionHeader.s_paddr );
		SectionHeader.s_vaddr = _htonl( SectionHeader.s_vaddr );
		SectionHeader.s_scnptr = _htonl( SectionHeader.s_scnptr );
		SectionHeader.s_type = _htons( SectionHeader.s_type );
		SectionHeader.s_relptr = _htonl( SectionHeader.s_relptr );
		SectionHeader.s_nreloc = _htons( SectionHeader.s_nreloc );
		DEBUG(( "Loading section %s", SectionHeader.s_name ));
		if ( SectionHeader.s_type & STYP_CLEAR ){
			if ( strcmp(SectionHeader.s_name, ".bss" ) == 0 )
				continue;
		}
		SectionInfo.SectionStart = SectionHeader.s_vaddr + LoadAddress;
		SectionInfo.SectionSize = SectionHeader.s_size;
		DEBUG(( "Writing section info" ));
		if ( WriteLink( TheLink, &SectionInfo, sizeof(SectionInfo), BOOT_TIMEOUT) <= 0 )
			ABORT(MISC_EXIT, (SE,"Boot: Could not write section info\n"));
		DEBUG(( "Wrote section info" ));
		fseek( fp, SectionHeader.s_scnptr, 0 );
		section_size = SectionHeader.s_size;
		StartOffset = EndOffset = SectionHeader.s_vaddr;
		relptr = SectionHeader.s_relptr;
		nreloc = SectionHeader.s_nreloc;
		while ( section_size > 0 ){
			buf_size = ( section_size > BOOT_BUFFER_SIZE )?BOOT_BUFFER_SIZE:section_size;
			EndOffset += buf_size;
			section_size -= buf_size;
			if (!( SectionHeader.s_type & STYP_CLEAR )){
				if ( fread((char *)BootBuffer,  sizeof(char), buf_size, fp ) != buf_size ){
					ABORT(MISC_EXIT, (SE,"Error in reading section %d (%s)", i, SectionHeader.s_name ));
				}
			}
			if (( nreloc > 0 )&&( strcmp( SectionHeader.s_name, ".data" ) == 0 ))
				Relocate( fp, StartOffset, EndOffset, &relptr, &nreloc, LoadAddress, TheLink, ServerInfo );
			DEBUG((  "Writing section %s size %d", SectionHeader.s_name, buf_size ));
			ops_res = WriteLink(TheLink, BootBuffer, (unsigned long) buf_size, BOOT_TIMEOUT);
			DEBUG((  "Wrote section %s size %d", SectionHeader.s_name, buf_size ));

			if (ops_res <= 0) {
				ABORT(MISC_EXIT, (SE,"Boot: write failed after %d bytes \n", buf_size));
			}
			StartOffset += buf_size;
		}
		fseek( fp, cur_pos, 0 );
	}
}
/*}}}*/

/*{{{  CoffLoad*/
/*  Load a COFF file at	the specified address
*/
/*
 *------------------------------------------------------------------------
 *
 *	CoffLoad :
 *		Loads the sections of a COFF file, starting at LoadAddress. 
 *		CoffParameters are sent first. Each section is preceded by 
 *		its SectionInfo structure.
 *	Parameters :
 *		CoffFileName : Name of COFF file
 *		LoadAddress : Address at which the file is to be loaded
 *			It is set to the next free address
 *		TheLink : Virtual/physical link number over which file is to
 *			be sent
 *		ServerInfo : structure containing information about server
 *
 *	Returns nothing.
 *
 *------------------------------------------------------------------------
 */
void CoffLoad( CoffFileName, TheLink, ServerInfo )
int TheLink;
char *CoffFileName;
SERVER_INFO *ServerInfo;
{
	FILE *CoffFile;
	int res;
	FILHDR FileHeader;
	AOUTHDR AoutHeader;
	unsigned long length;
	unsigned long temp;
	unsigned long LoadAddress;

	DEBUG((  "Coff file %s ", CoffFileName ));
	CoffFile = fopen( CoffFileName, "r" );
	if ( CoffFile == NULL ){
		ABORT(MISC_EXIT, (SE, "Boot: cannot find file \"%s\"", CoffFileName));
	}
	if (( res = fread( &FileHeader,  FILHSZ, 1, CoffFile )) != 1 ){
		ABORT(MISC_EXIT, (SE,  "%s: Could not read file header, res %d\n", CoffFileName, res ));
	}
	FileHeader.f_magic = _htons(FileHeader.f_magic);
	FileHeader.f_nscns = _htons(FileHeader.f_nscns);
	FileHeader.f_opthdr = _htons(FileHeader.f_opthdr);
	FileHeader.f_flags = _htons(FileHeader.f_flags);
	if (!ISACEMAGIC(FileHeader.f_magic)){
		ABORT(MISC_EXIT, (SE,  "%s: Bad magic %x\n", CoffFileName, FileHeader.f_magic ));
	}
	if (!( FileHeader.f_flags & F_EXEC )){
int TheLink;
		ABORT(MISC_EXIT, (SE,  "%s: Not an executable file flags %x\n" , CoffFileName, FileHeader.f_flags ));
	}
	if ( fread( &AoutHeader, AOUTHSZ, 1, CoffFile ) != 1 ){
		ABORT(MISC_EXIT, (SE,  "%s: Could not read optional header\n", CoffFileName ));
	}
	AoutHeader.entry = _htonl(AoutHeader.entry);
	CoffParams.Entry = AoutHeader.entry;
	CoffParams.NumSections = FileHeader.f_nscns;
	CoffParams.HeapSize = AppHeapSize;
	CoffParams.StackSize = AppStackSize;
	CoffParams.DebugFlag = (unsigned long)DebugApplicationSwitch;
	LoadSections( CoffFileName, CoffFile, (int)FileHeader.f_nscns, TheLink, ServerInfo );
	fclose( CoffFile );
	DEBUG(( "DebugSwitch %d Heap %d Stack %d",  DebugApplicationSwitch, AppHeapSize, AppStackSize ));
	if ( DebugApplicationSwitch ){
		DEBUG(( "Send debug command %s", DebuggerCommand ));
		length = strlen( DebuggerCommand )+1;
		if ((res =  WriteLink(TheLink, &length, sizeof(unsigned long), -1 /*BOOT_TIMEOUT*100*/)) < 0 )
			ABORT(MISC_EXIT, (SE,  "Could not send debug command len res %d len %d\n", res, length ))
		if ( WriteLink(TheLink, DebuggerCommand, length, BOOT_TIMEOUT*100) < 0 )
			ABORT(MISC_EXIT, (SE,  "Could not send debug command \n" ))
		DEBUG(( "Sent debug command %s", DebuggerCommand ));
	}
}
/*}}}*/
