/*{{{  includes and externs*/
/*
 --   ---------------------------------------------------------------------------
 --
 --      ISERVER  -  INMOS standard file server
 --
 --      iserver.c
 --
 --      The main body
 --
 --      Copyright (c) INMOS Ltd., 1988.
 --      All Rights Reserved.
 --
 --   ---------------------------------------------------------------------------
*/


#include <stdio.h>
#include <string.h>
#include <conc.h>

#include "inmos.h"
#include "isocket.h"
#include "vcpdefs.h"
#include "systype.h"
#include "queue.h"
#include "vcptype.h"
#include "tclVcpDefs.h"
#include "iserver.h"




#define T9_BOOT_FILE_ID	0x4c544207


EXTERN LINK OpenLink();
EXTERN int  CloseLink();
EXTERN int  ReadLink();
EXTERN int  WriteLink();
EXTERN int  ResetLink();
EXTERN int  AnalyseLink();
EXTERN int  TestError();
EXTERN int  TestRead();
EXTERN int  TestWrite();
/*}}}*/

/*{{{  statics*/
static SERVER_INFO *ServerInfoList = NULL;
static int NextServerId = 4;

PUBLIC BOOL AnalyseSwitch;                                                 /*  command line switches  */
PUBLIC BOOL TestErrorSwitch;
PUBLIC BOOL VerboseSwitch;
PUBLIC BOOL LinkSwitch;
PUBLIC BOOL ResetSwitch;
PUBLIC BOOL ServeSwitch;
PUBLIC BOOL LoadSwitch;
PUBLIC BOOL CocoPops;
PUBLIC BOOL OutSigio;

BYTE BootFileName[MAX_BOOT_FILE_LENGTH+1];

PRIVATE BYTE LinkName[MAX_LINK_NAME_LENGTH+1];
Semaphore ServerSemaphore = { 1, 0, 0};

extern char BootDeviceName[256];
char DeviceName[256];

PUBLIC BYTE *CoreDump;
PUBLIC INT32 CoreSize;

extern char *VCP_ErrorMessage[];
/*}}}*/

/*{{{  macros*/

/*	Get ServerInfo structure corresponding to a server
	identifier
*/
#define GET_SERVER_INFO( LinkId, ServerInfo ) {			\
	ServerInfo = ServerInfoList;				\
	while ( ServerInfo != NULL ){				\
		if ( ServerInfo->ServerId == LinkId )		\
			break;					\
		else						\
			ServerInfo = QNext(ServerInfo );	\
	}							\
}
/*}}}*/

/*{{{  ServerExit*/
/*	Close all file descriptors, delete the server
	information structure
*/
void ServerExit( SERVER_INFO *ServerInfo, int exitcode )	
{
	ServerInfo->TerminateFlag = TRUE;
	if (ServerInfo->StdOut > 0)fclose(ServerInfo->StdOutFP);
	if (ServerInfo->StdErr > 0)fclose(ServerInfo->StdErrFP);
	if ( ServerInfo->StdIn> 0 ) {
		fclose( ServerInfo->StdInFP );
		close( ServerInfo->StdIn );
	}
        CloseLink(ServerInfo->ServerId);
	QDelete( ServerInfoList, ServerInfo );
}
/*}}}*/

/*{{{  BootT9000*/
/*
 *   BootT9000  -  Boot a T9000 program
 */
PRIVATE VOID BootT9000(SERVER_INFO *ServerInfo)
{
	/*{{{  local variables*/
	FILE *Fd;
	BYTE *Buffer;
	int Length=0, Count=0;
	INT32 Size=0;
	static BYTE Flicks[]="|/-\\|/-\\" ;
	int Flick=0;
	int Loading =1;
	int TheLink;
	/*}}}*/

	Buffer = vcp_malloc( TRANSACTION_BUFFER_SIZE );
	if ( Buffer == NULL )
		ABORT(MISC_EXIT, (SE, "No memory for boot buffer"));

	/*{{{  Open file and process comment section*/
	if ( ( Fd=fopen( BootFileName, "rb" ) ) == NULL )
		ABORT(MISC_EXIT, (SE, "cannot find boot file \"%s\"", BootFileName));
	INFO(("Booting root transputer..."));
	if ( fread( &Length, sizeof(INT32), 1, Fd ) != 1 ) return;/* Skip ^GBTL */
	if ( Length != T9_BOOT_FILE_ID )
		ABORT( MISC_EXIT, (SE, "Illegal boot file" ));
	if ( fread( &Length, sizeof(INT32), 1, Fd ) != 1 ) return;/* Skip comment length */
	if ( ProcessComment( Fd ) < 0 )
		ABORT(MISC_EXIT, (SE, "Illegal boot file\n" ));
	DEBUG(( "Boot device %s", BootDeviceName ));
	DEBUG(( "Comment length %x", Length ));
	/*}}}*/
	if (( TheLink = OpenLink( BootDeviceName )) < 0 )
		ABORT(MISC_EXIT, (SE, "unable to access a transputer because\n%s",
		    VCP_ErrorMessage[-TheLink]));
	while ((Loading)&&(!ServerInfo->TerminateFlag))
	{
		/*{{{  read block size*/
		if ( fread( &Length, sizeof(INT32), 1, Fd ) != 1 ){
			Loading = 0; 
			break;
		}
		/*}}}*/
		/*{{{  read block*/
		if( Length > 0 ){
			if ( fread( Buffer, 1, Length, Fd ) != Length ){
				Loading = 0; 
				break;
			}
		}
		/*}}}*/
		/*{{{  send block to link*/
		DEBUG(( "Loading %d bytes", Length ));
		Count = WriteLink( TheLink, Buffer, Length, BOOT_TIMEOUT );
		Size += Count;
		if ( Count != Length )
		{
			if (Count < 0)
				ABORT(MISC_EXIT, (SE, "unable to write byte %d to the boot link because\n%s",
				    Size, VCP_ErrorMessage[0-Count]))
			    else
				ABORT(MISC_EXIT, (SE, "unable to write byte %d to the boot link", Size ));
		}
		/*}}}*/
                /*{{{  print /*/
                
#ifndef XCC     
                		if (isatty(1))
#endif          
                		{
                			INFO(( "%c\b", Flicks[Flick] ));
                			if ( ++Flick == 8 )
                				Flick -=8;
                		};
                /*}}}*/
	}
	fclose( Fd );
	CloseLink( TheLink );

	if ( isatty(1) )
	{
		INFO(( "\r                           \r" ));
	}
	else
	{
		INFO(("ok\n"));
	}
	DEBUG(( "BootT9000 over" ));
	DEBUG(( "booted %ld bytes", Size ));
}
/*}}}*/

/*{{{  BootT800*/
BootT800(int TheLink, SERVER_INFO *ServerInfo)
{
	FILE *Fd;
	BYTE Buffer[ BOOT_BUFFER_LENGTH ];
	int Length=0, Count=0;
	INT32 Size=0;
	static BYTE Flicks[]="|/-\\|/-\\" ;
	int Flick=0;

	if ( ( Fd=fopen( BootFileName, "rb" ) ) == NULL )
		ABORT(MISC_EXIT, (SE, "cannot find boot file \"%s\"", BootFileName));
	INFO(("Booting root transputer..."));
	while ( ( Length = fread( Buffer, 1, BOOT_BUFFER_LENGTH, Fd ) ) > 0 ) {
		DEBUG(("%d", Length));
		Count = WriteLink( TheLink, Buffer, Length, BOOT_TIMEOUT );
		Size += Count;
		if ( Count != Length ) {
			if (Count < 0)
				ABORT(MISC_EXIT, (SE, "unable to write byte %d to the boot link because\n%s",
                                             Size, VCP_ErrorMessage[0-Count]))
			else
				ABORT(MISC_EXIT, (SE, "unable to write byte %d to the boot link", Size ));
		}
#ifndef XCC
		if (isatty(1)) 
#endif
		{
			INFO(( "%c\b", Flicks[Flick] ));
			if ( ++Flick == 8 )
				Flick -=8;
		};
	}

	if ( isatty(1) )
	{
		INFO(( "\r                           \r" ));
	}
	else
	{
		INFO(("ok\n"));
	}
	DEBUG(( "booted %ld bytes", Size ));
}
/*}}}*/

/*{{{  PrintHelp*/

/*
 *   PrintHelp
 */


PRIVATE VOID PrintHelp ()
{
	fprintf( stdout, "\n%s : T9000 server\n", PROGRAM_NAME);
	fprintf( stdout, "Version %s.\n",  VERSION_NAME );
	fprintf( stdout, "Usage:  %s {%coption}\n\n", PROGRAM_NAME, SWITCH_CHAR );
	fprintf( stdout, "Options: \n" );
	fprintf( stdout, "        SB filename   boot the named file (same as %cSR%cSS%cSI%cSC filename)\n", SWITCH_CHAR,
	    SWITCH_CHAR, SWITCH_CHAR, SWITCH_CHAR );

	fprintf( stdout, "        SI            verbose mode\n");
	fprintf( stdout, "        SE            test the error flag\n");
	fprintf( stdout, "        SL linkname   use the named link\n");
	fprintf( stdout, "        SR            reset the root transputer\n");
	fprintf( stdout, "        SA            analyse and peek the root transputer\n");
	fprintf( stdout, "        SC filename   copy the named file to the link\n");
	fprintf( stdout, "        SP n          set peek size to n Kbytes.\n");
	fprintf( stdout, "        SS            serve the link\n\n");
}
/*}}}*/

/*{{{  ParseCommandLine*/
/*
 *   ParseCommandLine
 */

PRIVATE VOID ParseCommandLine (SERVER_INFO *ServerInfo)
{
	register BYTE *s, *t;
	int i;
	BOOL Gobble=TRUE;
	int TheLink = ServerInfo->ServerId;
	char temp[20];
	extern int DebugApplicationSwitch, AppHeapSize, AppStackSize;
	extern char DebuggerCommand[MAX_COMMAND_LINE_LENGTH];

	/*{{{  initialize all variables*/
	AnalyseSwitch = 0;
	TestErrorSwitch = 0;
	VerboseSwitch = 0;
	LinkSwitch = 0;
	ResetSwitch = 0;
	ServeSwitch = 0;
	LoadSwitch = 0;
	CocoPops = 0;
	CoreSize = DEFAULT_CORE_SIZE;
	
	BootFileName[0]=0;
	LinkName[0]=0;
	DeviceName[0] = '\0';
	
	s = ServerInfo->RealCommandLine;
	t = ServerInfo->DoctoredCommandLine;
	/*}}}*/

	/*{{{  skip to start of parameters*/
	while( (*s) && (*s != ' ') )                                            /* skip the first argv */
		++s;
	if ( *s == ' ' )
		++s;
	/*}}}*/

	for(;;)
	{
		if ( *s == '\0' ) {
			*t = *s;
			return;
		}
		/*{{{  Process -s options*/
		if (  (*s==SWITCH_CHAR) && ( (*(s+1)=='s') || (*(s+1)=='S') )  )
		{         /*  its a server option  */
			if (  strncmp(s, "-stack", 6 ) == 0 )
			{
				int i = 0;
				s+=6;
				while (*s==' ') s++;
				while ( (*s) && ( ( *s!=SWITCH_CHAR)  && (*s!=' ') ) )
					temp[i++]= *s++;
				temp[i] = '\0';
				i--;
				if (( temp[i] == 'k' )||( temp[i] == 'K' )){
					temp[i] = '\0';
					AppStackSize = atoi( temp ) * 1024;
				}
				else if (( temp[i] == 'm' )||( temp[i] == 'M' )){
					temp[i] = '\0';
					AppStackSize = atoi( temp ) * 1024 * 1024;
				}
				else {
					temp[i+1] = '\0';
					AppStackSize = atoi( temp );
				}
				continue;
			}
			s += 2;
			switch ( *s )
			{
			case 'a' :
			case 'A' :  
				++AnalyseSwitch; 
				++s; 
				break;
		
			case 'b' :
			case 'B' :
				++s; 
				++ResetSwitch; 
				++LoadSwitch; 
				++ServeSwitch; 
				++VerboseSwitch;
				while( *s == ' ' )
					++s;
				i = 0;
				while( (*s) && (*s != SWITCH_CHAR) && (*s != ' ') )
				{
					if ( i == MAX_BOOT_FILE_LENGTH )
						ABORT(MISC_EXIT, (SE, "boot filename is too long, maximum size is %d characters",
						    MAX_BOOT_FILE_LENGTH ));
					BootFileName[i++] = *s++;
				}
				if ( i == 0 )
					ABORT(MISC_EXIT, (SE, "expected a filename after %csb option", SWITCH_CHAR ));
				BootFileName[i] = 0;
				break;
		
			case 'c' :
			case 'C' :
				++s; 
				++LoadSwitch;
				while( *s == ' ' )
					++s;
				i = 0;
				while( (*s) && ( (*s!=SWITCH_CHAR) && (*s!=' ') ) )
				{
					if ( i == MAX_BOOT_FILE_LENGTH )
						ABORT(MISC_EXIT, (SE, "copy filename is too long, maximum size is %d characters",
						    MAX_BOOT_FILE_LENGTH ));
					BootFileName[i++] = *s++;
				}
				if ( i == 0 )
					ABORT(MISC_EXIT, (SE, "expected a filename after %csc option", SWITCH_CHAR ));
				BootFileName[i] = 0;
				break;
		
			case 'e' :
			case 'E' :  
				++s; 
				++TestErrorSwitch; 
				break;
		
			case 'i' :
			case 'I' :  
				++s; 
				++VerboseSwitch; 
				break;
		
			case 'l' :
			case 'L' :  
				++s; 
				++LinkSwitch;
				while( *s == ' ' )
					++s;
				i = 0;
				while( (*s) && ( (*s!=SWITCH_CHAR) && (*s!=' ') ) )
				{
					if ( i == MAX_LINK_NAME_LENGTH )
						ABORT(MISC_EXIT, (SE, "link name is too long, maximum size is %d characters",
						    MAX_LINK_NAME_LENGTH ));
					LinkName[i++] = *s++;
				}
				if ( i == 0 )
					ABORT(MISC_EXIT, (SE, "expected a name after %csl option", SWITCH_CHAR ));
				LinkName[i] = 0;
				break;
		
			case 'p' :
			case 'P' :  
				++s;
				while (*s == ' ')
					++s;
				CoreSize = (atoi(s) * 1024);
				if (CoreSize <= 0)
					ABORT(MISC_EXIT, (SE, "expected a positive number after %cSP option\n", SWITCH_CHAR));
				while( (*s) && (*s!=SWITCH_CHAR) )
					++s;
				break;
		
			case 'r' :
			case 'R' :  
				++s; 
				++ResetSwitch; 
				break;
		
			case 's' :
			case 'S' :  
				++s; 
				++ServeSwitch; 
				break;
		
			case 'z' :
			case 'Z' :  
				++s; 
				++CocoPops; 
				break;
		
			default  :
				*t++ = SWITCH_CHAR;
				*t++ = *(s-1);
				Gobble = FALSE;
				break;
			}
			if ( (Gobble == TRUE) && (*s == ' ') )
				++s;
		}
		/*}}}*/
		/*{{{  Process other server options*/
		else
		{
			if (  (*s==SWITCH_CHAR) && ( (*(s+1)=='p') || (*(s+1)=='P') )  )
			{
				int i=0;
				s+=2; /* skip -P or -p */
				while (*s==' ') s++;
				while ( (*s) && ( ( *s!=SWITCH_CHAR)  && (*s!=' ') ) )
					DeviceName[i++]= *s++;
				DeviceName[i] = '\0';
		
			}
			else if (  (*s==SWITCH_CHAR) && ( (*(s+1)=='d') || (*(s+1)=='D') )  )
			{
				int i = 0;
				s += 2;
				DebugApplicationSwitch = 1;
				while (*s==' ') s++;
				while ( (*s) && ( ( *s!=SWITCH_CHAR)  && (*s!=' ') ) )
					DebuggerCommand[i++]= *s++;
				DebuggerCommand[i] = '\0';
		
			}
			else if (  strncmp(s, "-heap", 5 ) == 0 )
			{
				int i = 0;
				s+=5;
				while (*s==' ') s++;
				while ( (*s) && ( ( *s!=SWITCH_CHAR)  && (*s!=' ') ) )
					temp[i++]= *s++;
				temp[i] = '\0';
				i--;
				if (( temp[i] == 'k' )||( temp[i] == 'K' )){
					temp[i] = '\0';
					AppHeapSize = atoi( temp ) * 1024;
				}
				else if (( temp[i] == 'm' )||( temp[i] == 'M' )){
					temp[i] = '\0';
					AppHeapSize = atoi( temp ) * 1024 * 1024;
				}
				else {
					temp[i+1] = '\0';
					AppHeapSize = atoi( temp );
				}
			}
			else
			{
				*t++ = *s++;
			}
		}
		/*}}}*/
	}
}
/*}}}*/

/*{{{  ProcessTBuf*/
/*	Execute a serve request. Returns True if successful, and
	false otherwise
*/
Boolean ProcessTBuf( SERVER_INFO *ServerInfo, int *Count )
{
	/*{{{  local variables*/
	BYTE *SendBuf;
	int SockId;
	int Length, i;
	FILE *SockFP;
	Boolean ret = False;
	BYTE *RecvBuf = ServerInfo->RecvBuffer;
	/*}}}*/

	SendBuf = ServerInfo->SendBuffer;
	if (( SendBuf[0] == SP_SOCKMODE1 )||
	    ( SendBuf[0] == SP_SOCKMODE2 ))
		DEBUG(( "Tag %d:%d", SendBuf[0], SendBuf[1] ))
	else
		DEBUG(( "Tag %d", SendBuf[0] ))
	if ( SendBuf[0] == SP_EXIT ){
		/*{{{  set terminate flag, return success*/
		ServerInfo->TerminateFlag = TRUE;
		*Count = 1; 
		ServerInfo->RecvBuffer[0] = SP_SUCCESS;
		if ( ServerInfo->StdIn > 0 ) close( ServerInfo->StdIn );
		return(True);
		/*}}}*/
	}
	if ( SendBuf[0] == SP_COMMAND ){
		/*{{{  return command line*/
		RecvBuf[0] = SP_SUCCESS;
		if ( SendBuf[1] ){
			strcpy( &(RecvBuf[3]), ServerInfo->RealCommandLine );
			Length = strlen( ServerInfo->RealCommandLine );
		}
		else {
			strcpy( &(RecvBuf[3]), ServerInfo->DoctoredCommandLine );
			Length = strlen( ServerInfo->RealCommandLine );
		}
		RecvBuf[1] = Length%256;
		RecvBuf[2] = Length/256;
		*Count = Length+3;
		return(True);
		/*}}}*/
	}
	if (( SendBuf[0] == SP_SOCKMODE1 )&&( SendBuf[1] == SP_SOCKCLOSE )){
		/* Donot close std files */
		READ_INT32( &(SendBuf[2]), SockId );
		if (( SockId >=0 )&&( SockId <= 2 )){
			*Count = 1; 
			ServerInfo->RecvBuffer[0] = SP_SUCCESS;
			return( True );
		}
	}
	if ( SendBuf[0] == SP_CLOSE ){
		/* Donot close std files */
		READ_INT32( &(SendBuf[1]), SockId );
		if (( SockId >=0 )&&( SockId <= 2 )){
			*Count = 1; 
			ServerInfo->RecvBuffer[0] = SP_SUCCESS;
			return( True );
		}
	}
	if (( ServerInfo->StdIn < 0 )&&( ServerInfo->StdOut < 0 )&&
	    ( ServerInfo->StdErr < 0 ))
		return(False);
	if (( SendBuf[0] == SP_SOCKMODE1 )||( SendBuf[0] == SP_SOCKMODE2 )){
		/*{{{  replace socket id for stdin, stdout and stderr*/
		if (( SendBuf[1] == SP_FCNTL )||
		    ( SendBuf[1] == SP_IOCTL )||
		    ( SendBuf[1] == SP_SOCKREAD )||
		    ( SendBuf[1] == SP_SOCKREADV )||
		    ( SendBuf[1] == SP_SOCKWRITE )||
		    ( SendBuf[1] == SP_SOCKWRITEV )||
		    ( SendBuf[1] == SP_SOCKSEEK )){
			READ_INT32( &(SendBuf[2]), SockId );
			if (( SockId == 0 )&&( ServerInfo->StdIn > 0 )){
				WRITE_INT32( &(SendBuf[2]), ServerInfo->StdIn )
			}
			else if (( SockId == 1 )&&( ServerInfo->StdOut > 0 )){
				WRITE_INT32( &(SendBuf[2]), ServerInfo->StdOut )
			}
			else if (( SockId == 2 )&&( ServerInfo->StdErr > 0 )){
				WRITE_INT32( &(SendBuf[2]), ServerInfo->StdErr )
			}
			else return(False);
		}
		return(False);
		/*}}}*/
	}
	if (( SendBuf[0] == SP_CLOSE )||
	    ( SendBuf[0] == SP_READ )||
	    ( SendBuf[0] == SP_WRITE )||
	    ( SendBuf[0] == SP_GETBLOCK )||
	    ( SendBuf[0] == SP_PUTBLOCK )||
	    ( SendBuf[0] == SP_GETS )||
	    ( SendBuf[0] == SP_PUTS )||
	    ( SendBuf[0] == SP_FLUSH )||
	    ( SendBuf[0] == SP_SEEK )||
	    ( SendBuf[0] == SP_TELL )||
	    ( SendBuf[0] == SP_EOF )||
	    ( SendBuf[0] == SP_ERROR )){
			/*{{{  perform file operations for stdin, stdout and stderr*/
			/*{{{  get corresponding file pointer*/
			READ_INT32( &(SendBuf[1]), SockId );
			if (( SockId == 0 )&&( ServerInfo->StdIn > 0 )){
				SockFP = ServerInfo->StdInFP;
			}
			else if (( SockId == 1 )&&( ServerInfo->StdOut > 0 )){
				SockFP =  ServerInfo->StdOutFP;
			}
			else if (( SockId == 2 )&&( ServerInfo->StdErr > 0 )){
				SockFP =  ServerInfo->StdErrFP;
			}
			else return(False);
			/*}}}*/
			switch ( SendBuf[0] ){
			case SP_PUTS :
				/*{{{  write to new file pointer*/
				{
					Length = SendBuf[5] + SendBuf[6] * 256;
					SendBuf[7+Length] = '\n';
					fwrite( &(SendBuf[7]), sizeof(char), Length+1, SockFP );
					if ( SockId == 1 )fflush( SockFP );
					*Count = 1;
					RecvBuf[0] = SP_SUCCESS;
					ret = True; 
					break;
				}
				/*}}}*/
			case SP_WRITE :
			case SP_PUTBLOCK:
				/*{{{  write to new file pointer*/
				{
					Length = SendBuf[5] + SendBuf[6] * 256;
					Length = fwrite( &(SendBuf[7]), sizeof(char), Length, SockFP );
					if ( SockId == 1 )fflush( SockFP );
					*Count = 3;
					RecvBuf[0] = SP_SUCCESS;
					RecvBuf[1] = Length%256;
					RecvBuf[2] = Length/256;
					ret = True; 
					break;
				}
				/*}}}*/
			case SP_GETS:
				/*{{{  read from new file pointer*/
				{
					Length = SendBuf[5] + SendBuf[6] * 256;
					fgets( &(RecvBuf[3]), Length, SockFP );
					Length = strlen( &(RecvBuf[3]));
					if ( RecvBuf[3+Length-1] == '\n' ){
						RecvBuf[3+Length] = 0;
						--Length;
					}
					RecvBuf[0] = SP_SUCCESS;
					RecvBuf[1] = Length%256;
					RecvBuf[2] = Length/256;
					ret = True; 
					break;
				}
				/*}}}*/
			case SP_READ:
			case SP_GETBLOCK:
				/*{{{  read from new file pointer*/
				{
					Length = SendBuf[5] + SendBuf[6] * 256;
					Length = fread( &(RecvBuf[3]), sizeof(char), Length, SockFP );
					RecvBuf[0] = SP_SUCCESS;
					RecvBuf[1] = Length%256;
					RecvBuf[2] = Length/256;
					ret = True; 
					break;
				}
				/*}}}*/
			case SP_CLOSE :
				/*{{{  will be closed later by ServerExit*/
				*Count = 1;
				RecvBuf[0] = SP_SUCCESS;
				return(True);
				/*}}}*/
			default :
				break;
			}
			/*}}}*/
		}
		return(ret);
	}
	/*}}}*/

/*{{{  BlockedServeT9000*/

/*
 *   ServeT9000  -  the main loop (read buffer from link, call a function, write buffer to link)
 */
PRIVATE int BlockedServeT9000( int ServerId, SERVER_INFO *ServerInfo )
{
	int Count;
	int length, res;
	int TheLink = ServerInfo->ServerId;

	while ( !ServerInfo->TerminateFlag ){
		Count = ReadLink(ServerId, &(ServerInfo->SendBuffer[0]), TRANSACTION_BUFFER_SIZE, -1);
		if ( Count > 0 ){
			if( !ProcessTBuf( ServerInfo, &Count ) ){
				Count = server_transaction( ServerInfo->SendBuffer, Count, ServerInfo->RecvBuffer );
			}
			if ( WriteLink(ServerId, ServerInfo->RecvBuffer, Count, TIMEOUT ) < 0 )
				ABORT(MISC_EXIT, (SE,"WriteLink timed out\n" ));
		}
	}
	ServerExit(ServerInfo, 0);
}
/*}}}*/

/*{{{  ServerThread*/

void ServeT9000_Thread(int FIXME, int ServerId, SERVER_INFO *ServerInfo, char *BoundCommand)
{
	int Count;
	int length, res;
	int TheLink = ServerInfo->ServerId;

	while ( !ServerInfo->TerminateFlag ){
		Count = ReadLink(ServerId, &(ServerInfo->SendBuffer[0]), TRANSACTION_BUFFER_SIZE, -1);
		SemWait( &ServerSemaphore );
		if ( Count > 0 ){
			if( !ProcessTBuf( ServerInfo, &Count ) )
				Count = server_transaction( ServerInfo->SendBuffer, Count, ServerInfo->RecvBuffer );
			if ( WriteLink(ServerId, ServerInfo->RecvBuffer, Count, TIMEOUT*10 ) < 0 )
				ABORT(MISC_EXIT, (SE,"%d:WriteLink timed out Count %d Tag %d:%d\n", ServerInfo->ServerId, Count, (unsigned char)ServerInfo->SendBuffer[0], (unsigned char)ServerInfo->SendBuffer[1] ));
		}
		SemSignal( &ServerSemaphore );
	}
	QueueTclCommand( BoundCommand );
	SemWait( &ServerSemaphore );
	ServerExit(ServerInfo, 0);
	SemSignal( &ServerSemaphore );
}
/*}}}*/

/*{{{  UnBlockedServeT9000*/
PRIVATE int UnBlockedServeT9000( int ServerId, SERVER_INFO *ServerInfo, char *BoundCommand)
{
	Process *pid;
	pid = ProcAlloc( ServeT9000_Thread, DEFAULT_STACKSIZE, 3, ServerId, ServerInfo, BoundCommand );
	if ( pid != NULL )
		ProcRunHigh( pid );
	else 
		fprintf( stderr, "Could not create server thread\n" );
DEBUG(( "Unblocked serve T9" ));
return( 0 );
}
/*}}}*/

/*{{{  GetRequest*/

/* GetRequest -- Get a request from the link; handle timeouts */
int GetRequest( SERVER_INFO *ServerInfo, BYTE *ServBuf )
{
	register int   Count,Size;
	BYTE Tbuf[2];
	int TheLink = ServerInfo->ServerId;

	while (((Count = ReadLink(TheLink, &Tbuf[0], 1, TIMEOUT)) == VCP_TimeOut) || (Count == 0)) {
		DEBUG(("ReadLink retrying"));
	}

	/* We've got the first byte of a request, now try for the rest of the length */
	Count = ReadLink(TheLink, &Tbuf[1], 1, TIMEOUT);

	if (Count < 0)
		ABORT(MISC_EXIT, (SE, "Unable to get request from link because\n%s", VCP_ErrorMessage[0-Count]))

	Size = Tbuf[0] + (Tbuf[1]*256);

	if (Size < 6)
		ABORT(MISC_EXIT, (SE, "protocol error, packet size is too small (%d)", Size));

	if (Size > (TRANSACTION_BUFFER_SIZE - 2))
		ABORT(MISC_EXIT, (SE, "protocol error, packet size is too large (%d)", Size));

	Count = ReadLink(TheLink, ServBuf, Size, TIMEOUT);

	if (Count < 0)
		ABORT(MISC_EXIT, (SE, "Unable to get request body from link because\n%s", VCP_ErrorMessage[0-Count]))

	if( Count != Size )
		ABORT(MISC_EXIT, (SE, "protocol error, timed out getting the request body"));

	return Size;
}
/*}}}*/

/*{{{  BlockedServeT800*/

/*
 *   ServeT800  -  the main loop (read buffer from link, call a function, write buffer to link)
 */
PRIVATE int BlockedServeT800( int TheLink, SERVER_INFO *ServerInfo )
{
	int Count;
	BYTE Tbuf[2];

	while ( !ServerInfo->TerminateFlag ){
		Count = GetRequest(ServerInfo, &(ServerInfo->SendBuffer[0]));
		if ( Count > 0 ){
			if( !ProcessTBuf( ServerInfo, &Count ) ){
				Count = server_transaction( ServerInfo->SendBuffer, Count, ServerInfo->RecvBuffer );
			}
			Tbuf[0] = Count % 256; Tbuf[1] = Count / 256;
			if ( WriteLink(TheLink, Tbuf, 2, TIMEOUT ) < 0 )
				ABORT(MISC_EXIT, (SE,"WriteLink timed out\n" ));
			if ( WriteLink(TheLink, ServerInfo->RecvBuffer, Count, TIMEOUT ) < 0 )
				ABORT(MISC_EXIT, (SE,"WriteLink timed out\n" ));
		}
	}
	ServerExit(ServerInfo, 0);
}
/*}}}*/

/*{{{  ServerThread*/

void ServeT800_Thread(int FIXME, int TheLink, SERVER_INFO *ServerInfo, char *BoundCommand)
{
	int Count;
	BYTE Tbuf[2];

	while ( !ServerInfo->TerminateFlag ){
		Count = GetRequest(ServerInfo, &(ServerInfo->SendBuffer[0]));
		SemWait( &ServerSemaphore );
		if ( Count > 0 ){
			if( !ProcessTBuf( ServerInfo, &Count ) )
				Count = server_transaction( ServerInfo->SendBuffer, Count, ServerInfo->RecvBuffer );
			Tbuf[0] = Count % 256; Tbuf[1] = Count / 256;
			if ( WriteLink(TheLink, Tbuf, 2, TIMEOUT ) < 0 )
				ABORT(MISC_EXIT, (SE,"WriteLink timed out\n" ));
			if ( WriteLink(TheLink, ServerInfo->RecvBuffer, Count, TIMEOUT ) < 0 )
				ABORT(MISC_EXIT, (SE,"WriteLink timed out\n" ));
		}
		SemSignal( &ServerSemaphore );
	}
	QueueTclCommand( BoundCommand );
	SemWait( &ServerSemaphore );
	ServerExit(ServerInfo, 0);
	SemSignal( &ServerSemaphore );
}
/*}}}*/

/*{{{  UnBlockedServeT800*/
PRIVATE int UnBlockedServeT800( int TheLink, SERVER_INFO *ServerInfo, char *BoundCommand)
{
	Process *pid;

	pid = ProcAlloc( ServeT800_Thread, DEFAULT_STACKSIZE, 3, TheLink, ServerInfo, BoundCommand );
	if ( pid != NULL )
		ProcRunHigh( pid );
	else 
		fprintf( stderr, "Could not create server thread\n" );
	return( 0 );
}
/*}}}*/


/*{{{  QuotedArg*/

PRIVATE int QuotedArg(str)
char *str;
{
	while (*str) {
		if ((*str == ' ') || (*str == '\t'))
			return TRUE;

		str++;
	}

	return FALSE;
}
/*}}}*/

/*{{{  ServerInit*/
/*
 *   ServerInit - main program of iserver
 */
PUBLIC int ServerInit (ServerName, StdIn, StdOut, StdErr, BoundCommand, LinkInfo, argc, argv)
char *ServerName;
char *LinkInfo;
int argc;
char **argv;
int StdIn, StdOut, StdErr;
char *BoundCommand;

{
	/*{{{  local variables*/
	int Quoted;
	int ExitStatus = MISC_EXIT;
	int LinkStatus;
	BYTE *c, *ALinkName;
	SERVER_INFO *ServerInfo;
	int ServerId;
	int TheLink;
	
	/*}}}*/
	
	if ( argc == 1 )
	{
		/*{{{  PrintHelp*/
		PrintHelp();
		if ( BoundCommand )QueueTclCommand(BoundCommand);
		return ( MISC_EXIT );
		/*}}}*/
	}
	/*{{{  Initialise ServerInfo structure*/
	ServerInfo = (SERVER_INFO *)vcp_malloc( sizeof(SERVER_INFO) );
	if ( ServerInfo == NULL )
		ABORT(MISC_EXIT, (SE, "Out of memory") )
	EnQueue( ServerInfoList, ServerInfo );
	ServerInfo->ServerId = NextServerId++;
	ServerInfo->TerminateFlag = False;
	ServerInfo->StdIn = StdIn; 
	ServerInfo->StdOut = StdOut; 
	ServerInfo->StdErr = StdErr;
	if ( StdIn > 0 )ServerInfo->StdInFP = fdopen( StdIn, "r" );
	if ( StdOut > 0 )ServerInfo->StdOutFP = fdopen( StdOut, "a" );
	if ( StdErr > 0 )ServerInfo->StdErrFP = fdopen( StdErr, "a" );
	/*}}}*/

	/*{{{  ParseCommandLine*/
	*(ServerInfo->RealCommandLine) = 0;
	*(ServerInfo->DoctoredCommandLine) = 0;
	while ( argc-- > 0 )
	{
		if ( ( MAX_COMMAND_LINE_LENGTH - strlen( ServerInfo->RealCommandLine) ) < strlen( *argv ) )
		{
			ABORT(MISC_EXIT, (SE, "Command line too long (at \"%s\")", *argv) )
		}
		if (QuotedArg(*argv)) {
			Quoted = TRUE;
			(void)strcat( ServerInfo->RealCommandLine, "\"" );
		}
		else
			Quoted = FALSE;
	
		(void)strcat( ServerInfo->RealCommandLine, *argv );
		if (Quoted)
			(void)strcat( ServerInfo->RealCommandLine, "\"");
		(void)strcat( ServerInfo->RealCommandLine, " " );
		++argv;
	}
	for ( c=ServerInfo->RealCommandLine; *c ; ++c )                                     /*  strip of the last space  */
		;
	if ( *--c == ' ' )
		*c = 0;
	
	ParseCommandLine(ServerInfo);
	DEBUG(("original line is \"%s\"", ServerInfo->RealCommandLine));
	DEBUG(("doctored line is \"%s\"", ServerInfo->DoctoredCommandLine));
	/*}}}*/

	/*{{{  Get device name, print debug info*/
	if ( CocoPops )
		printf("(analyse=%d error=%d verbose=%d link=%d reset=%d serve=%d load=%d coco=%d)\n", AnalyseSwitch, TestErrorSwitch,
		    VerboseSwitch, LinkSwitch, ResetSwitch, ServeSwitch, LoadSwitch, CocoPops );
	
	DEBUG(("using \"%s\" as the boot file", BootFileName));
	strcpy( DeviceName, LinkInfo );
	DEBUG(("and \"%s\" as the link name", DeviceName));
	/*}}}*/

	/*{{{  OpenLink*/
	if ( ( ServerId = OpenLink( DeviceName ) ) < 0 )
		ABORT(MISC_EXIT, (SE, "unable to access a transputer because\n%s",
		    VCP_ErrorMessage[-ServerId]));
	TheLink = ServerInfo->ServerId = ServerId;
	/*}}}*/

	if ( ResetSwitch && AnalyseSwitch )
		ABORT(MISC_EXIT, (SE, "reset and analyse are incompatible"));

	if ( ResetSwitch )
	{
		/*{{{  Reset*/
		LinkStatus = ResetLink(ServerId);
		if (LinkStatus != 1)
			ABORT(MISC_EXIT, (SE, "failed to reset root transputer because\n%s",
			    VCP_ErrorMessage[0-LinkStatus]));
		DEBUG(("reset root"));
		/*}}}*/
	}

	if ( AnalyseSwitch )
	{
		/*{{{  Analyse*/
		LinkStatus = AnalyseLink(ServerId);
		if (LinkStatus != 1)
			ABORT(MISC_EXIT, (SE, "failed to analyse root transputer because\n%s",
			    VCP_ErrorMessage[0-LinkStatus]));
		DEBUG(("analyse root"));
		/*}}}*/
	}


	/*{{{  Boot */
	if ( LoadSwitch ){
		if ( strcmp( ServerName, COFF_SERV_INIT ) == 0 ) {
			if ( strlen(LinkInfo) == 2 )
				CoffBootT800(ServerId, ServerInfo, BootFileName);
			else {
				CoffBootT9000(ServerInfo, BootFileName);
				DEBUG(( "Coff Boot T9 over" ));
				if ( ServerInfo->TerminateFlag == TRUE ){
					if ( BoundCommand )QueueTclCommand(BoundCommand);
					return(MISC_EXIT);
				}
#if 0
				/*{{{  OpenLink*/
				if ( ( ServerId = OpenLink( DeviceName ) ) < 0 )
					ABORT(MISC_EXIT, (SE, "unable to access a transputer because\n%s",
					    VCP_ErrorMessage[-ServerId]));
				TheLink = ServerInfo->ServerId = ServerId;
				DEBUG(( "Coff server link %d", TheLink ));
#endif
				/*}}}*/
 
			}
		}
		else {
			if ( strlen(LinkInfo) == 2 )
				BootT800(ServerId, ServerInfo);
			else {
				CloseLink( ServerId );
				BootT9000(ServerInfo);
				if ( ServerInfo->TerminateFlag == TRUE ){
					if ( BoundCommand )QueueTclCommand(BoundCommand);
					return(MISC_EXIT);
				}
				/*{{{  OpenLink*/
				if ( ( ServerId = OpenLink( DeviceName ) ) < 0 )
					ABORT(MISC_EXIT, (SE, "unable to access a transputer because\n%s",
					    VCP_ErrorMessage[-ServerId]));
				TheLink = ServerInfo->ServerId = ServerId;
				/*}}}*/
 
			}
		}
	}
	/*}}}*/

	if ( ServeSwitch ){
		if ( strlen(LinkInfo) == 2){
			/*{{{  ServeT800*/
			if ( BoundCommand != NULL ){
				ExitStatus = UnBlockedServeT800(ServerId, ServerInfo, BoundCommand);
			}
			else {
				ExitStatus = BlockedServeT800(ServerId, ServerInfo);
			}
			/*}}}*/
		}
		else {
			/*{{{  ServeT9000*/
			if ( BoundCommand != NULL ){
				ExitStatus = UnBlockedServeT9000(ServerId, ServerInfo, BoundCommand);
			}
			else {
				ExitStatus = BlockedServeT9000(ServerId, ServerInfo);
			}
			/*}}}*/
		}
	}
	else if ( ResetSwitch || AnalyseSwitch || LoadSwitch || TestErrorSwitch )
		ExitStatus = MISC_EXIT;
	else
	{
		INFO(("(no action taken)\n"));
	}
	return( ServerId );
}
/*}}}*/

/*{{{  ServerQuit*/
ServerQuit( ServerId )
int ServerId;
{
	int LinkStatus;
	SERVER_INFO *ServerInfo;

	GET_SERVER_INFO( ServerId, ServerInfo );
	if ( ServerInfo != NULL )
		ServerExit( ServerInfo, 0 );
	return( 0 );

}
/*}}}*/

/*{{{  SetServerSemaphore*/
SetServerSemaphore()
{
	SemWait( &ServerSemaphore );	/* Called by TCL init */
}
/*}}}*/

/*
 *   Eof
 */
